# Relion star table names for particle coordinates
STAR_X_COLUMN_NAME = "rlnCoordinateX"
STAR_Y_COLUMN_NAME = "rlnCoordinateY"
STAR_Z_COLUMN_NAME = "rlnCoordinateZ"
STAR_ROTATION_0 = "rlnAngleRot"
STAR_ROTATION_1 = "rlnAngleTilt"
STAR_ROTATION_2 = "rlnAnglePsi"


# column names for napari points layer for oriented particles
NAPARI_NORMAL_0 = "normal-0"
NAPARI_NORMAL_1 = "normal-1"
NAPARI_NORMAL_2 = "normal-2"
NAPARI_UP_0 = "up-0"
NAPARI_UP_1 = "up-1"
NAPARI_UP_2 = "up-2"
ROTATION = "rotation"
